# Copyright 2024 The jaxcam Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for converting cameras to and from ray-based representations."""

from typing import Any, Optional
from flax import struct
import jax
import jax.numpy as jnp
from jaxcam._src import camera as jaxcam
from jaxcam._src import math


@struct.dataclass
class Rays:
  """Represents cameras as a bundle of rays.

  Supports rays represented as directions and origins or directions and moments
  (Plücker coordinates [1]). Whereas a line can be represented by any origin
  along the ray, it is represented by a unique (normalized) direction and
  moment.

  This class should not be instantiated directly. Use the `create()` function
  instead.

  [1]: https://en.wikipedia.org/wiki/Pl%C3%BCcker_coordinates

  Attributes:
    directions: (..., 3) Ray directions.
    origins: (..., 3) Ray origins.
    moments: (..., 3) Plücker moments.
    shape: (int, ...) The shape of the rays.
  """

  directions: jnp.ndarray
  origins: jnp.ndarray

  @classmethod
  def create(
      cls,
      directions: jnp.ndarray,
      *,
      origins: Optional[jnp.ndarray] = None,
      moments: Optional[jnp.ndarray] = None,
  ) -> 'Rays':
    """Creates a Rays object.

    Note that if the rays are instantiated with moments, the origins could be
    any point along the ray. In other words,
    `Rays.create(rays.directions, moments=rays.moments).origins` is not
    guaranteed to be the same as `rays.origins`.

    Args:
      directions: (..., 3) Unit normalized ray directions.
      origins: (..., 3) Ray origins.
      moments: (..., 3) Plücker moments.

    Returns:
      A Rays object.
    """
    # Ensure directions are unit normalized.
    norms = jnp.linalg.norm(directions, axis=-1, keepdims=True)
    directions = directions / norms
    if origins is not None and moments is not None:
      raise ValueError('Cannot specify both origins and moments.')
    if origins is None and moments is None:
      raise ValueError('Must specify either origins or moments.')

    if moments is not None:
      if directions.shape != moments.shape:
        raise ValueError(
            'Directions and moments must have the same shape. Got '
            f'{directions.shape} and {moments.shape}'
        )
      origins = jnp.cross(directions, moments, axis=-1)
    elif origins is not None:
      if directions.shape != origins.shape:
        raise ValueError(
            'Directions and origins must have the same shape. Got '
            f'{directions.shape} and {origins.shape}'
        )
    return cls(directions=directions, origins=origins)

  @property
  def moments(self) -> jnp.ndarray:
    norm = jnp.linalg.norm(self.directions, axis=-1, keepdims=True)
    return jnp.cross(self.origins, self.directions / norm, axis=-1)

  def to_raymap_6d(self, use_plucker: bool = False) -> jnp.ndarray:
    """Returns a 6D raymap representation of the rays.

    Raymap represents rays as [moments, directions] (Plücker coordinates) or
    [origins, directions]

    Args:
      use_plucker: If True, uses Plücker coordinates.

    Returns:
      A (..., 6) raymap representation of the rays.
    """
    if use_plucker:
      return jnp.concatenate((self.moments, self.directions), axis=-1)
    else:
      return jnp.concatenate((self.origins, self.directions), axis=-1)

  @classmethod
  def from_raymap(
      cls,
      raymap: jnp.ndarray,
      use_plucker: bool = False,
  ) -> 'Rays':
    """Creates a Rays object from a raymap representation."""
    moments_or_origins = raymap[..., :3]
    directions = raymap[..., 3:]
    if use_plucker:
      moments, origins = moments_or_origins, None
    else:
      moments, origins = None, moments_or_origins
    return cls.create(
        directions=directions,
        moments=moments,
        origins=origins,
    )

  @property
  def shape(self) -> tuple[int, ...]:
    return self.directions.shape


def get_rays_from_camera(
    camera: jaxcam.Camera,
    normalize: bool = True,
) -> Rays:
  """Computes rays unprojected from every pixel.

  Currently only supports a single camera.

  Args:
    camera: Camera to unproject.
    normalize (bool): If True, normalizes the directions to have unit norm.

  Returns:
    A tuple of (directions, origins) where both are (H, W, 3).
  """
  pixels = jaxcam.get_pixel_centers(*camera.image_size)
  directions = jaxcam.pixels_to_rays(camera, pixels, normalize=normalize)
  origins = jnp.tile(
      camera.position, (int(camera.image_size[1]), int(camera.image_size[0]), 1)
  )
  return Rays.create(directions=directions, origins=origins)


def get_camera_from_rays(
    rays: Rays,
    use_ransac: bool = False,
    ransac_parameters: Optional[dict[str, Any]] = None,
) -> jaxcam.Camera:
  """Recovers a Camera from a ordered grid of rays.

  Performs a least-squares fit for a pinhole camera using DLT. See Sec 4.1 of
  Hartley and Zisserman [1] for more details. Note that the returned camera may
  have parameters that are slightly different from the original camera due to
  numerical precision.

  [1]
  https://github.com/DeepRobot2020/books/blob/master/Multiple%20View%20Geometry%20in%20Computer%20Vision%20(Second%20Edition).pdf

  Args:
    rays: Camera rays where directions and origins are (H, W, 3).
    use_ransac: If True, uses RANSAC to robustly estimate the camera parameters.
    ransac_parameters: Parameters for RANSAC. See
      _compute_homography_transform_ransac for valid keys.

  Returns:
    camera: jaxcam.Camera.
  """
  position = jnp.mean(rays.origins, axis=(0, 1))
  height, width, _ = rays.directions.shape
  identity_camera = jaxcam.Camera.create(
      orientation=jnp.eye(3),
      image_size=jnp.array([width, height]),
      position=jnp.zeros(3),
      focal_length=jnp.array([1.0]),
      principal_point=jnp.array([0, 0]),
  )
  identity_rays = get_rays_from_camera(identity_camera)
  intrinsics, orientation, _ = ray_dlt(
      identity_rays.directions.reshape(-1, 3),
      rays.directions.reshape(-1, 3),
      use_ransac=use_ransac,
      ransac_parameters=ransac_parameters,
  )
  focal_length = (intrinsics[0, 0] + intrinsics[1, 1]) / 2
  principal_point = intrinsics[:2, 2]
  skew = intrinsics[0, 1]
  return jaxcam.Camera.create(
      orientation=orientation,
      image_size=jnp.array([width, height]),
      position=position,
      focal_length=focal_length,
      principal_point=principal_point,
      skew=skew,
  )


def ray_dlt(
    directions1,
    directions2,
    use_ransac: bool = False,
    ransac_parameters: Optional[dict[str, Any]] = None,
) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
  """Solves for the optimal K and R transformation between two sets of rays.

  X2 ~ H @ X1 = R.T @ K_inv @ X1

  Args:
    directions1: (N, 3)
    directions2: (N, 3)
    use_ransac: If True, uses RANSAC to robustly estimate the camera parameters.
    ransac_parameters: Parameters for RANSAC. See
      _compute_homography_transform_ransac for valid keys.

  Returns:
    A tuple of (calibration, rotation, homography) where the calibration
    matrix (3, 3) is upper triangular, the rotation matrix (3, 3) is in
    world2camera convention, and the homography matrix (3, 3) is the optimal
    homography transform between the two sets of rays.
  """
  directions1 = directions1 / jnp.linalg.norm(
      directions1, axis=-1, keepdims=True
  )
  directions2 = directions2 / jnp.linalg.norm(
      directions2, axis=-1, keepdims=True
  )
  if use_ransac:
    if ransac_parameters is None:
      ransac_parameters = {}
    if 'rng' not in ransac_parameters:
      ransac_parameters['rng'] = jax.random.PRNGKey(0)
    homography = _compute_homography_transform_ransac(
        directions1=directions1,
        directions2=directions2,
        **ransac_parameters,
    )
  else:
    homography = _compute_homography_transform(directions1, directions2)
  out = jnp.linalg.qr(homography)
  # H = R.T @ K_inv, so R.T corresponds to the Q (orthonormal) and K_inv to the
  # R (upper triangular) in QR decomposition.
  intrinsics = jnp.linalg.inv(out.R)
  rotation = out.Q.T
  intrinsics /= intrinsics[2, 2]
  intrinsics, rotation = _positivize(intrinsics, rotation)
  return intrinsics, rotation, homography


@jax.jit
def _compute_homography_transform_ransac(
    rng,
    directions1: jnp.ndarray,
    directions2: jnp.ndarray,
    num_iterations: int = 200,
    num_correspondences: int = 8,
    threshold_deg: float = 1.0,
) -> jnp.ndarray:
  """Solves for the optimal homography transform with RANSAC.

  Args:
    rng: Random Number Generator.
    directions1: (N, 3) Unit normalized ray directions.
    directions2: (N, 3) Unit normalized ray directions.
    num_iterations: Number of iterations to run.
    num_correspondences: Number of correspondences to use per iteration.
    threshold_deg: Threshold (in degrees) for inliers.

  Returns:
    A (3, 3) homography matrix that has the most inliers.
  """

  def ransac_body(_, state):
    rng, best_inliers, best_homography = state
    rng, key = jax.random.split(rng)
    random_inds = jax.random.choice(
        key,
        jnp.arange(directions1.shape[0]),
        shape=(num_correspondences,),
        replace=False,
    )
    homography = _compute_homography_transform(
        directions1[random_inds], directions2[random_inds]
    )
    # Compute inliers by checking alignment of projected ray (opposite rays
    # also count as inliers).
    proj = math.matmul(homography, directions1.T).T
    proj = proj / jnp.linalg.norm(proj, keepdims=True, axis=-1)
    dot = jnp.clip(jnp.sum(proj * directions2, axis=-1), -1, 1)
    dist = jnp.arccos(jnp.abs(dot))
    is_inlier = dist * 180 / jnp.pi < threshold_deg
    num_inliers = jnp.sum(is_inlier)
    best_homography, best_inliers = jax.lax.cond(
        num_inliers > best_inliers,
        lambda _: (homography, num_inliers),
        lambda _: (best_homography, best_inliers),
        operand=None,
    )
    return rng, best_inliers, best_homography

  state = (rng, -1, jnp.eye(3))
  _, _, best_homography = jax.lax.fori_loop(
      0, num_iterations, ransac_body, state
  )
  return best_homography


@jax.jit
def _compute_homography_transform(
    directions1: jnp.ndarray,
    directions2: jnp.ndarray,
) -> jnp.ndarray:
  """Solves for the optimal homography transform between two sets of rays.

  Finds the homography transform H:
  x2 ~ H @ x1

  Args:
    directions1 (jnp.ndarray): (N, 3).
    directions2 (jnp.ndarray): (N, 3).

  Returns:
    A (3, 3) homography matrix.
  """
  x1, y1, z1 = jnp.split(directions1, 3, axis=-1)
  x2, y2, z2 = jnp.split(directions2, 3, axis=-1)
  z = jnp.zeros_like(x1)
  # Eq 4.1 in H&Z.
  a_x = jnp.concatenate(
      (z2 * x1, z2 * y1, z2 * z1, z, z, z, -x2 * x1, -x2 * y1, -x2 * z1), axis=1
  )
  a_y = jnp.concatenate(
      (z, z, z, -z2 * x1, -z2 * y1, -z2 * z1, y2 * x1, y2 * y1, y2 * z1), axis=1
  )
  homogeneous_matrix = jnp.concatenate((a_x, a_y), axis=0)
  _, _, vh = jnp.linalg.svd(homogeneous_matrix, full_matrices=False)
  homography = vh[-1]
  # Homography matrix is only defined up to a scale factor:
  homography = homography / jnp.linalg.norm(homography)
  homography = homography.reshape(3, 3)
  return homography


def _positivize(
    calib: jnp.ndarray, rotation: jnp.ndarray
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Makes diagonal of calibration matrix positive.

  Ensures the diagonal entries of the calibration matrix (i.e. focal lengths)
  are positive, flipping the sign of the rows of the rotation matrix as
  necessary. Also ensures that the rotation matrix is valid (det(R) == 1).

  Args:
    calib: (3, 3) Calibration matrix.
    rotation: (3, 3) Rotation matrix.

  Returns:
    A tuple of (calib, rotation).
  """
  sign = jnp.sign(calib.diagonal())
  calib = calib * sign[None, :]
  rotation = rotation * sign[:, None]
  # Ensure valid rotation matrix.
  rotation = jax.lax.cond(
      jnp.linalg.det(rotation) < 0, lambda: -rotation, lambda: rotation
  )
  return calib, rotation



# Copyright 2024 The jaxcam Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Math utilities."""

import jax
from jax import numpy as jnp


def matmul(a: jnp.ndarray, b: jnp.ndarray) -> jnp.ndarray:
  """jnp.matmul uses bfloat16 by default on TPU, this prevents that.

  References:
    https://github.com/google/jax#current-gotchas
    https://github.com/google/jax/issues/2161
    https://github.com/google/jax/issues/7010

  Args:
    a: the left matrix of a matmul.
    b: the right matrix of a matmul.

  Returns:
    The result of the matmul.
  """
  return jnp.matmul(a, b, precision=jax.lax.Precision.HIGHEST)


def einsum(*args, **kwargs) -> jnp.ndarray:
  """jnp.einsum uses bfloat16 by default on TPU, this prevents that."""
  return jnp.einsum(*args, **kwargs, precision=jax.lax.Precision.HIGHEST)


def skew(vector: jnp.ndarray) -> jnp.ndarray:
  """Builds a skew matrix ("cross product matrix") for a vector.

  References:
    https://en.wikipedia.org/wiki/Skew-symmetric_matrix.

  Args:
    vector: (3,) A 3-vector.

  Returns:
    A (3, 3) skew-symmetric matrix such that W @ v == w x v where w is the input
    vector, W is the corresponding skew-symmatrix matrix, and v is another
    vector.
  """
  vector = jnp.reshape(vector, (3))
  return jnp.array([
      [0.0, -vector[2], vector[1]],
      [vector[2], 0.0, -vector[0]],
      [-vector[1], vector[0], 0.0],
  ])


def transform_point(
    point: jnp.ndarray,
    scale: jnp.ndarray,
    rotation: jnp.ndarray,
    translation: jnp.ndarray,
) -> jnp.ndarray:
  """Transforms the given point x as `scale * rotation @ x + translation`."""
  return scale * matmul(rotation, point) + translation



# Copyright 2024 The jaxcam Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for reading and writing Cameras."""

import json
import pathlib
from typing import Union

from etils import epath
import jaxcam
import numpy as np

PathType = Union[str, pathlib.PurePath]


def from_nerfies_json_file(path: PathType) -> jaxcam.Camera:
  """Load a camera from a JSON file in the Nerfies format.

  This format directly serializes the fields of the camera dataclass, and
  additionally an `image_size` field containing `(image_size_x, image_size_y)`.

  This format is used by the Nerfies datasets, which is available publicly at
  https://github.com/google/nerfies#camera

  Args:
    path: the path of the file to read.

  Returns:
    The loaded camera instance.
  """
  path = epath.Path(path)
  with path.open('r') as fp:
    camera_dict = json.load(fp)
  camera_dict = {k: np.asarray(v) for k, v in camera_dict.items()}
  return jaxcam.Camera.create(**camera_dict)


def to_nerfies_json_file(path: PathType, camera: jaxcam.Camera) -> None:
  """Saves a camera to a JSON file in the Nerfies format.

  Please see :obj:`from_nerfies_json_file` for details.

  Args:
    path: the path to save to.
    camera: the camera to save.
  """
  path = epath.Path(path)
  camera_dict = {
      k: v.tolist() if hasattr(v, 'tolist') else v
      for k, v in camera.pytree_dict().items()
  }
  with path.open('w') as fp:
    json.dump(camera_dict, fp, indent=2)



# Copyright 2024 The jaxcam Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=g-multiple-import,g-importing-member,useless-import-alias
"""The main package namespace."""

# A new PyPI release will be pushed everytime `__version__` is increased
# When changing this, also update the CHANGELOG.md
__version__ = '0.3.0'

# Note: import <name> as <name> is required for names to be exported.
# See PEP 484 & https://github.com/google/jax/issues/7570
from jaxcam._src.camera import (
    Camera as Camera,
    concatenate as concatenate,
    create as create,
    crop as crop,
    depth_to_ray_depth as depth_to_ray_depth,
    essential_matrix as essential_matrix,
    fundamental_matrix as fundamental_matrix,
    get_pixel_centers as get_pixel_centers,
    invert_radial_distortion_coefficients as invert_radial_distortion_coefficients,
    local_points_to_world_points as local_points_to_world_points,
    look_at as look_at,
    pixels_to_local_rays as pixels_to_local_rays,
    pixels_to_points as pixels_to_points,
    pixels_to_rays as pixels_to_rays,
    project as project,
    ray_depth_to_depth as ray_depth_to_depth,
    relative_transform as relative_transform,
    scale as scale,
    transform as transform,
    update_intrinsic_matrix as update_intrinsic_matrix,
    update_translation as update_translation,
    update_world_to_camera_matrix as update_world_to_camera_matrix,
    world_points_to_local_points as world_points_to_local_points,
    world_to_camera_matrix as world_to_camera_matrix,
)
from jaxcam._src.rays import (
    get_camera_from_rays as get_camera_from_rays,
    get_rays_from_camera as get_rays_from_camera,
    Rays as Rays,
)



# Copyright 2024 The jaxcam Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Camera transformation utilities.
"""

from typing import Tuple
import jax
import jax.numpy as jnp
import jaxcam
from jaxcam._src import math
import optax


def rts_to_sim3(
    rotation: jnp.ndarray, translation: jnp.ndarray, scale: float
) -> jnp.ndarray:
  """Converts a rotation, translation and scale to a homogeneous transform.

  Args:
    rotation: (3, 3) An orthonormal rotation matrix.
    translation: (3,) A 3-vector representing a translation.
    scale: A scalar factor.

  Returns:
    (4, 4) A homogeneous transformation matrix.
  """

  transform = jnp.eye(4)
  transform = transform.at[:3, :3].set(rotation * scale)
  transform = transform.at[:3, 3].set(translation)

  return transform


def sim3_to_rts(
    transform: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
  """Converts a homogeneous transform to rotation, translation and scale.

  Args:
    transform: (4, 4) A homogeneous transformation matrix.

  Returns:
    rotation: (3, 3) An orthonormal rotation matrix.
    translation: (3,) A 3-vector representing a translation.
    scale: A scalar factor.
  """

  eps = jnp.float32(jnp.finfo(jnp.float32).tiny)
  rotation_scale = transform[..., :3, :3]
  # Assumes rotation is an orthonormal transform, thus taking norm of first row.
  scale = optax.safe_norm(rotation_scale, min_norm=eps, axis=1)[0]
  rotation = rotation_scale / scale
  translation = transform[..., :3, 3]
  return rotation, translation, scale


def relativize_cameras(
    reference: jaxcam.Camera, target: jaxcam.Camera
) -> tuple[jaxcam.Camera, jaxcam.Camera]:
  """Transforms the target camera to be in the reference camera's frame.

  Relativizes the camera such that the target cameras are in the reference
  camera's reference frame. I.e., the reference camera's camera coordinate frame
  becomes the world coordinate frame.

  Args:
    reference: The reference camera.
    target: The target camera.

  Returns:
    A tuple containing the transformed reference and target cameras. The
    reference camera will have an identity world-to-camera matrix.
  """

  transform = jaxcam.relative_transform(reference, target)
  rel_reference = jaxcam.update_world_to_camera_matrix(reference, jnp.eye(4))
  rel_target = jaxcam.update_world_to_camera_matrix(target, transform)
  return rel_reference, rel_target


def transform_camera(
    camera: jaxcam.Camera, transform: jax.Array
) -> jaxcam.Camera:
  """Transform cameras.

  Args:
    camera: The camera to be transformed.
    transform: (4, 4) Transformation matrix comprised of rotation, translation
      and scale.

  Returns:
    Transformed cameras.
  """
  rotation, translation, scale = sim3_to_rts(transform)
  transform = rts_to_sim3(rotation, translation, 1.0)
  camera_from_world = jaxcam.world_to_camera_matrix(camera)

  camera_from_transform = math.matmul(
      camera_from_world, transform[None]
  ).squeeze(0)
  camera = jaxcam.update_world_to_camera_matrix(camera, camera_from_transform)
  camera = jaxcam.scale(camera, scale)

  return camera


def transform_to_identity_rotation(
    camera: jaxcam.Camera,
) -> jaxcam.Camera:
  """Transforms a camera to have an identity rotation."""
  rotation = camera.orientation.T
  transform = rts_to_sim3(rotation, jnp.zeros(3), 1.0)
  return transform_camera(camera, transform)


def relativize_rotation(
    reference: jaxcam.Camera, target: jaxcam.Camera
) -> tuple[jaxcam.Camera, jaxcam.Camera]:
  """Transforms the target camera to be in the reference camera's frame.

  Relativizes the camera such that the target cameras are in the reference
  camera's reference frame. I.e., the reference camera's camera coordinate frame
  becomes the world coordinate frame.

  Args:
    reference: The reference camera.
    target: The target camera.

  Returns:
    A tuple containing the transformed reference and target cameras. The
    reference camera will have an identity world-to-camera matrix.
  """

  transform = reference.orientation.T
  transform = rts_to_sim3(transform, jnp.zeros(3), 1.0)
  rel_reference = transform_camera(reference, transform)
  rel_target = transform_camera(target, transform)
  return rel_reference, rel_target



# Copyright 2024 The jaxcam Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A JAX-based camera class."""

from collections.abc import Sequence
import dataclasses
import enum
import functools
from typing import Any, Generator, Optional, Union
import warnings

from flax import struct
import jax
import jax.numpy as jnp
from jaxcam._src import math
from numpy import typing as npt


def _camera_field(unbatched_shape: tuple[int, ...], **kwargs) -> ...:
  return dataclasses.field(
      metadata={
          'pytree_node': True,
          'unbatched_shape': unbatched_shape,
      },
      **kwargs,
  )


class ProjectionType(enum.Enum):
  """Camera projection type (perspective pinhole, or fisheye)."""

  PERSPECTIVE = 'perspective'
  FISHEYE = 'fisheye'


@struct.dataclass
class Camera:
  """The base class for a pinhole camera model.

  This class should not be instantiated directly. Use the `create()` function
  instead.

  References:
    https://en.wikipedia.org/wiki/Distortion_(optics)#Software_correction
    https://web.archive.org/web/20180312205006/https://www.asprs.org/wp-content/uploads/pers/1966journal/may/1966_may_444-462.pdf

  Attributes:
    orientation: (..., 3, 3) array of world-to-camera rotations.
    position: (..., 3) array of world-space positions.
    focal_length: (...) array or float of focal length in x.
    principal_point: (..., 2) array of the image-space principal point.
    image_size: (..., 2) array of image sizes.
    skew: (...) array or float of skew coefficients.
    pixel_aspect_ratio: (...) array or float of pixel aspect ratios (height /
      width)
    radial_distortion: (..., 4) array of radial distortion coefficients.
    tangential_distortion: (..., 2) array of tangential distortion coefficients.
    scale_factor_x: convenience property for focal_length.
    scale_factor_y: convenience property for focal_length * pixel_aspect_ratio.
    principal_point_x: convenience property for principal_point(..., 0).
    principal_point_y: convenience property for principal_point(..., 1).
    image_size_x: convenience property for image_size(..., 0).
    image_size_y: convenience property for image_size(..., 1).
    optical_axis: convenience property for orientation(..., 2, :).
    translation: the camera translation vector.
    world_to_camera_matrix: (..., 4, 4) the world-to-camera matrix of the camera
      (also known as the extrinsic matrix).
    intrinsic_matrix: (..., 3, 3) the intrinsic matrix of the camera.
    projection_type: the projection type of the camera.
    has_distortion: will be True if either radial or tangential distortion is
      enabled.
    has_radial_distortion: will be True if radial distortion is enabled.
    has_tangential_distortion: will be True if tangential distortion is enabled.
    use_inverted_distortion: if True, the meaning of distort/undistort is
      flipped.
  """
  orientation: jnp.ndarray = _camera_field((3, 3))
  position: jnp.ndarray = _camera_field((3,))
  focal_length: Union[jnp.ndarray, float] = _camera_field(())
  principal_point: jnp.ndarray = _camera_field((2,))
  image_size: jnp.ndarray = _camera_field((2,))
  skew: Union[jnp.ndarray, float] = _camera_field(())
  pixel_aspect_ratio: Union[jnp.ndarray, float] = _camera_field(())
  radial_distortion: jnp.ndarray | None = _camera_field((4,), default=None)
  tangential_distortion: jnp.ndarray | None = _camera_field((2,), default=None)

  # Non-pytree fields. For example, options to control camera logic. Logic is
  # baked in when jitting, so these fields do not need to be passed into jitted
  # functions and so can be left out of the pytree representation of the camera.
  projection_type: ProjectionType = struct.field(
      pytree_node=False, default=ProjectionType.PERSPECTIVE
  )
  use_inverted_distortion: bool = struct.field(pytree_node=False, default=False)

  @classmethod
  def create(
      cls,
      orientation: Optional[jnp.ndarray] = None,
      position: Optional[jnp.ndarray] = None,
      focal_length: Optional[jnp.ndarray] = None,
      principal_point: Optional[jnp.ndarray] = None,
      image_size: Optional[jnp.ndarray] = None,
      skew: Union[jnp.ndarray, float] = 0.0,
      pixel_aspect_ratio: Union[jnp.ndarray, float] = 1.0,
      radial_distortion: Optional[jnp.ndarray] = None,
      tangential_distortion: Optional[jnp.ndarray] = None,
      invert_distortion: bool = False,
      is_fisheye: bool = False,
  ) -> 'Camera':
    """Creates a camera with reasonable default values."""
    if position is None:
      position = jnp.zeros(3)
    if orientation is None:
      orientation = jnp.eye(3)
    if image_size is None:
      image_size = jnp.ones(2)
    if principal_point is None:
      principal_point = image_size / 2.0
    if focal_length is None:
      # Default focal length produces a FoV of 2*atan(0.5) ~= 53 degrees.
      focal_length = image_size[..., 0]

    # Ensure all items are strongly typed arrays to avoid triggering a cache
    # miss during JIT compilation due to weak type semantics.
    # See: https://jax.readthedocs.io/en/latest/type_promotion.html
    asarray = functools.partial(jnp.asarray, dtype=jnp.float32)

    kwargs = {
        'orientation': asarray(orientation),
        'position': asarray(position),
        'focal_length': asarray(focal_length),
        'principal_point': asarray(principal_point),
        'image_size': asarray(image_size),
        'skew': asarray(skew),
        'pixel_aspect_ratio': asarray(pixel_aspect_ratio),
    }

    if radial_distortion is not None:
      # Insert the 4th radial distortion coefficient if not present.
      radial_distortion = jnp.pad(
          asarray(radial_distortion),
          pad_width=(0, 4 - radial_distortion.shape[-1]),
      )
      kwargs['radial_distortion'] = asarray(radial_distortion)

    if tangential_distortion is not None:
      kwargs['tangential_distortion'] = asarray(tangential_distortion)

    if invert_distortion and (
        radial_distortion is not None or tangential_distortion is not None
    ):
      kwargs['use_inverted_distortion'] = True

    if is_fisheye:
      kwargs['projection_type'] = ProjectionType.FISHEYE
    else:
      kwargs['projection_type'] = ProjectionType.PERSPECTIVE

    return cls(**kwargs)

  def batch_axes(self) -> Optional['Camera']:
    """Returns the batch axis for each field.

    Camera fields may be individually batched or unbatched. This function
    inspects the field shapes and returns a Camera with an axis for each
    field: 0 if the field is batched, or None if the field is unbatched.
    The output of this function may be used as part of the in_axes parameter
    to jax.vmap. Using Camera with labels as fields is a bit of an abuse of the
    dataclass but jax.vmap requires that the structure of the input and in_axes
    matches exactly.

    Returns:
      a Camera with batch axis labels in each field, or None if no fields are
      batched.
    """
    fields = {f.name: f for f in self.pytree_fields()}
    field_is_batched = {
        k: len(jnp.array(v).shape) > len(fields[k].metadata['unbatched_shape'])
        for k, v in self.pytree_dict().items()
    }
    if not any(field_is_batched.values()):
      return None
    field_axes = {k: 0 if v else None for k, v in field_is_batched.items()}
    return type(self)(**field_axes)

  @classmethod
  def pytree_fields(cls) -> list[dataclasses.Field[Any]]:
    fields = dataclasses.fields(cls)
    return [f for f in fields if f.metadata['pytree_node']]

  def pytree_dict(self) -> dict[str, Any]:
    """Returns a dict of the dataclass containing only pytree fields.

    Returns:
      A dictionary containing enabled pytree fields. None-type fields will be
      omitted.
    """
    pytree_fields = self.pytree_fields()
    return {
        field.name: self.__dict__[field.name]
        for field in pytree_fields
        if self.__dict__[field.name] is not None
    }

  @property
  def ndim(self) -> int:
    """Returns the number of batch dimensions."""
    return len(self.shape)

  @property
  def shape(self) -> tuple[int, ...]:
    """The shape of the batch dimensions of the camera."""
    batch_axes = self.batch_axes()
    if batch_axes is None:
      return ()

    shape = ()
    # This class assumes that the batch dimensions of all batched Attributes
    # are the same, so we can just return the shape of any such attribute.
    for field in dataclasses.fields(self):
      batch_axis = getattr(batch_axes, field.name)
      if batch_axis is not None:
        array = getattr(self, field.name)
        unbatched_ndim = len(field.metadata['unbatched_shape'])
        batch_ndim = array.ndim - unbatched_ndim
        shape = array.shape[:batch_ndim]
        break

    return shape

  def __len__(self) -> int:
    """Return the number of cameras in the batch.

    Raises:
      TypeError: If the camera is unbatched.

    Returns:
      The number of cameras in the batch. Follows the same convention as
        numpy as returns the first dimension of the shape if multiple batch
        dimensions exist.
    """
    if self.ndim == 0:
      raise TypeError('len() of unbatched camera')
    return self.shape[0]

  def __getitem__(self, index: npt.ArrayLike) -> 'Camera':
    """Convenience method for slicing the camera.

    Slices the camera. Any slicing object that works with numpy arrays will
    work here.

    All batched fields must have the same batch dimension and unbatched fields
    will remain unbatched.

    Examples:

    ```python
    # Get the first 10 cameras.
    cameras[:10]

    # Get the 5th camera.
    cameras[5]

    # Index with an array.
    cameras[jnp.ndarray([0, 2, 4, 6])]
    ```

    Args:
      index: The slicing object.

    Raises:
      IndexError: If called on an unbatched camera.

    Returns:
      The sliced camera.
    """
    batch_axes = self.batch_axes()
    if batch_axes is None:
      raise IndexError('Cannot slice unbatched camera.')

    kwargs = {}
    pytree_dict = self.pytree_dict()
    for k, v in self.__dict__.items():
      if k in pytree_dict and v is not None:
        kwargs[k] = v[index] if getattr(batch_axes, k) is not None else v
      else:
        kwargs[k] = v

    return self.__class__(**kwargs)

  def __iter__(self) -> Generator['Camera', None, None]:
    """Iterate over the camera batch.

    Iterates along the first batch dimension of the camera. For example, the
    following are equivalent:

      ```python
      [cam for cam in camera]
      [camera[i] for i in len(camera)]
      ```

    Raises:
      IndexError: If called on an unbatched camera.

    Yields:
      Cameras with each element sliced along the first batch dimension.
    """
    if self.ndim == 0:
      raise TypeError('iteration over a 0-d camera')
    for i in range(len(self)):
      yield self[i]

  @property
  def scale_factor_x(self) -> jnp.ndarray:
    return self.focal_length  # pytype: disable=bad-return-type  # jax-ndarray

  @property
  def scale_factor_y(self) -> jnp.ndarray:
    return self.focal_length * self.pixel_aspect_ratio  # pytype: disable=bad-return-type  # jax-ndarray

  @property
  def principal_point_x(self) -> jnp.ndarray:
    return self.principal_point[..., 0]

  @property
  def principal_point_y(self) -> jnp.ndarray:
    return self.principal_point[..., 1]

  @property
  def image_size_x(self) -> jnp.ndarray:
    return self.image_size[..., 0]

  @property
  def image_size_y(self) -> jnp.ndarray:
    return self.image_size[..., 1]

  @property
  def optical_axis(self) -> jnp.ndarray:
    return self.orientation[..., 2, :]

  @property
  def has_distortion(self) -> bool:
    return self.has_radial_distortion or self.has_tangential_distortion

  @property
  def has_radial_distortion(self) -> bool:
    return self.radial_distortion is not None

  @property
  def has_tangential_distortion(self) -> bool:
    return self.tangential_distortion is not None

  @property
  def translation(self) -> jnp.ndarray:
    # pylint: disable=invalid-unary-operand-type
    return math.matmul(-self.orientation, self.position[..., None]).squeeze(-1)

  @property
  def world_to_camera_matrix(self) -> jnp.ndarray:
    """Returns the 4x4 matrix that takes world points to camera coordinates."""
    matrix = jnp.empty((*self.shape, 3, 4))
    matrix = matrix.at[..., :3, :3].set(self.orientation)
    matrix = matrix.at[..., :3, 3].set(self.translation)
    return _to_matrix_4x4(matrix)

  @property
  def camera_to_world_matrix(self) -> jnp.ndarray:
    """Returns the 4x4 matrix that takes camera points to world coordinates."""
    matrix = jnp.empty((*self.shape, 3, 4))
    matrix = matrix.at[..., :3, :3].set(self.orientation.transpose(-1, -2))
    matrix = matrix.at[..., :3, 3].set(self.position)
    return _to_matrix_4x4(matrix)

  @property
  def intrinsic_matrix(self) -> jnp.ndarray:
    """Returns the intrinsic matrix that maps local coordinates to pixels."""
    matrix = jnp.zeros((*self.shape, 3, 3))
    matrix = matrix.at[..., 0, 0].set(self.scale_factor_x)
    matrix = matrix.at[..., 0, 1].set(self.skew)
    matrix = matrix.at[..., 0, 2].set(self.principal_point_x)
    matrix = matrix.at[..., 1, 1].set(self.scale_factor_y)
    matrix = matrix.at[..., 1, 2].set(self.principal_point_y)
    matrix = matrix.at[..., 2, 2].set(1)
    return matrix


def create(*args, **kwargs) -> Camera:
  warnings.warn(
      'The `create()` function has been moved to `Camera.create()`.'
      '`create()` will be removed in the future.',
      DeprecationWarning,
  )
  return Camera.create(*args, **kwargs)


def _to_matrix_4x4(matrix):
  """Converts a matrix to a 4x4 matrix."""
  if matrix.shape[-2] > 4 or matrix.shape[-1] > 4:
    raise ValueError('Matrix dimensions must be a maximum of 4x4.')

  num_rows, num_cols = matrix.shape[-2:]
  matrix_4x4 = jnp.zeros((*matrix.shape[:-2], 4, 4))
  matrix_4x4 = matrix_4x4.at[..., 3, 3].set(1)
  matrix_4x4 = matrix_4x4.at[..., :num_rows, :num_cols].set(matrix)
  return matrix_4x4


def _distort_local_pixels(
    camera: Camera, x: jnp.ndarray, y: jnp.ndarray
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Distorts normalized image pixels."""
  if camera.use_inverted_distortion:
    x, y = _radial_and_tangential_undistort(
        x,
        y,
        radial_distortion=camera.radial_distortion,
        tangential_distortion=camera.tangential_distortion,
    )
  elif camera.has_distortion:
    x, y = _radial_and_tangential_distort(
        x,
        y,
        radial_distortion=camera.radial_distortion,
        tangential_distortion=camera.tangential_distortion,
    )

  return x, y


def _undistort_local_pixels(
    camera: Camera, x: jnp.ndarray, y: jnp.ndarray
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Undistorts normalized image pixels."""
  if camera.use_inverted_distortion:
    x, y = _radial_and_tangential_distort(
        x,
        y,
        radial_distortion=camera.radial_distortion,
        tangential_distortion=camera.tangential_distortion,
    )
  elif camera.has_distortion:
    x, y = _radial_and_tangential_undistort(
        x,
        y,
        radial_distortion=camera.radial_distortion,
        tangential_distortion=camera.tangential_distortion,
    )

  return x, y


def pixels_to_local_rays(
    camera: Camera, pixels: jnp.ndarray, normalize: bool = True
) -> jnp.ndarray:
  """Returns local ray directions for the provided pixels."""
  y = ((pixels[..., 1] - camera.principal_point_y) / camera.scale_factor_y)
  x = (
      pixels[..., 0] - camera.principal_point_x - y * camera.skew
  ) / camera.scale_factor_x

  x, y = _undistort_local_pixels(camera, x, y)

  dirs = jnp.stack([x, y, jnp.ones_like(x)], axis=-1)
  if camera.projection_type is ProjectionType.FISHEYE:
    theta = jnp.sqrt(x ** 2 + y ** 2)
    theta = jnp.minimum(jnp.pi, theta)
    sin_theta_over_theta = jnp.sin(theta) / theta
    fisheye_dir = jnp.stack([
        dirs[..., 0] * sin_theta_over_theta,
        dirs[..., 1] * sin_theta_over_theta,
        jnp.cos(theta),
    ], axis=-1)
    eps = jnp.finfo(x.dtype).eps
    # It is approximately perspective when the viewing ray is too close to the
    # optical axis.
    if normalize:
      dirs = dirs / jnp.linalg.norm(dirs, axis=-1, keepdims=True)
    dirs = jnp.where(theta[..., None] < eps, dirs, fisheye_dir)
  else:
    if normalize:
      dirs = dirs / jnp.linalg.norm(dirs, axis=-1, keepdims=True)

  return dirs


def pixels_to_rays(
    camera: Camera, pixels: jnp.ndarray, normalize: bool = True
) -> jnp.ndarray:
  """Returns world-space rays for the provided pixels.

  Args:
    camera: Camera instance.
    pixels: (..., 2) array containing 2d pixel positions.
    normalize: If True, normalize the output ray directions. If False, the
      returned directions will be points at the image plane at distance 1 from
      camera center. Defaults to True.

  Returns:
      An array containing the normalized ray directions in world coordinates.
  """
  if pixels.shape[-1] != 2:
    raise ValueError('The last dimension of pixels must be 2.')

  batch_shape = pixels.shape[:-1]
  pixels = jnp.reshape(pixels, (-1, 2))

  local_rays_dir = pixels_to_local_rays(camera, pixels, normalize=normalize)
  rays_dir = math.matmul(
      jnp.swapaxes(camera.orientation, -1, -2), local_rays_dir[..., jnp.newaxis]
  )
  rays_dir = jnp.squeeze(rays_dir, axis=-1)
  rays_dir = rays_dir.reshape((*batch_shape, 3))
  return rays_dir


def pixels_to_points(
    camera: Camera, pixels: jnp.ndarray, depth: jnp.ndarray
) -> jnp.ndarray:
  """Unprojects pixels and depth to (x,y,z,w) homogenous world-space points."""
  rays_through_pixels = pixels_to_rays(camera, pixels)
  ray_depth = depth_to_ray_depth(camera, rays_through_pixels, depth)
  points = rays_through_pixels * ray_depth[..., None] + camera.position
  pad_shape = [(0, 0)] * len(depth.shape) + [(0, 1)]
  points = jnp.pad(points, pad_shape, constant_values=1.0)
  return points


def world_points_to_local_points(camera: Camera,
                                 world_points: jnp.ndarray) -> jnp.ndarray:
  """Transforms world-space (x,y,z) points to local-space (x,y,z) points.

  Local-space coordinates are also known as camera-space coordinates.

  Args:
    camera: The camera defining the local coordinate frame.
    world_points: (..., 3) The world-space points to transform.

  Returns:
    The local-space coordinates.
  """
  world_points_flat = world_points.reshape((-1, 3))
  translated_points = world_points_flat - camera.position
  local_points = (math.matmul(camera.orientation, translated_points.T)).T
  return local_points.reshape(world_points.shape)


def local_points_to_world_points(
    camera: Camera, local_points: jnp.ndarray
) -> jnp.ndarray:
  """Transforms local-space (x,y,z) points to world-space (x,y,z) points.

  Local-space coordinates are also known as camera-space coordinates.

  Args:
    camera: The camera defining the local coordinate frame.
    local_points: (..., 3) The local-space points to transform.

  Returns:
    The world-space coordinates.
  """
  local_points_flat = local_points.reshape((-1, 3))

  rotated_points = math.matmul(camera.orientation.T, local_points_flat.T).T
  world_points = rotated_points + camera.position
  return world_points.reshape(local_points.shape)


def depth_to_ray_depth(camera: Camera, ray: jnp.ndarray,
                       depth: jnp.ndarray) -> jnp.ndarray:
  """Converts depth along the optical axis to depth along ray.

  Args:
    camera: The camera defining the local coordinate frame.
    ray: (..., 3) The ray corresponding to each point in `depth`.
    depth: (...,) The depth along the optical axis for each ray in `ray`.

  Returns:
    (..., 3) The depth along the optical axis for each ray.
  """
  cosa = math.matmul(ray, camera.optical_axis)
  return depth / cosa


def ray_depth_to_depth(camera: Camera, ray: jnp.ndarray,
                       ray_depth: jnp.ndarray) -> jnp.ndarray:
  """Converts depth along ray to depth along the optical axis.

  Args:
    camera: The camera defining the local coordinate frame.
    ray: (..., 3) The ray corresponding to each point in `ray_depth`.
    ray_depth: (...,) The depth along the ray for each ray in `ray`.

  Returns:
    (..., 3) The depth along the optical axis for each ray.
  """
  cosa = math.matmul(ray, camera.optical_axis)
  return ray_depth * cosa


def world_to_camera_matrix(camera: Camera) -> jnp.ndarray:
  """Returns the 4x4 matrix that takes world points to camera coordinates.

  DEPRECATED: Use Camera.world_to_camera_matrix.

  Args:
    camera: The camera to query.

  Returns:
    The world-to-camera matrix.
  """
  return camera.world_to_camera_matrix


def update_world_to_camera_matrix(
    camera: Camera, matrix: jnp.ndarray
) -> Camera:
  """Sets the transform that takes world points to camera coordinates.

  Args:
    camera: Camera to be used to set camera_from_world transform.
    matrix: A (..., 4, 4) jax array containing a homogeneous transform, where
      the last row is [0, 0, 0, 1].

  Returns:
    Camera with updated orientation and position after applying transform.
  """

  if matrix.shape[-2:] != (4, 4):
    raise ValueError('Last two axes of transform must have shape (3, 3).')

  orientation = matrix[..., :3, :3]
  translation = matrix[..., :3, 3]
  position = math.einsum('... j, ... j k -> ... k', translation, -orientation)
  return camera.replace(position=position, orientation=orientation)


def update_translation(camera: Camera, translation: jnp.ndarray) -> Camera:
  """Updates the translation of the camera.

  Args:
    camera: Camera to set the translation for.
    translation: A (..., 3) array containing the new translation.

  Returns:
    Camera with updated translation.
  """
  new_position = math.einsum(
      '... j, ... j k -> ... k', translation, -camera.orientation
  )
  return camera.replace(position=new_position)


def update_intrinsic_matrix(
    camera: Camera, intrinsic_matrix: jnp.ndarray
) -> Camera:
  """Sets the camera intrinsics based on an intrinsic matrix.

  Args:
    camera: Camera to set the intrinsics for.
    intrinsic_matrix: A (..., 3, 3) jax array containing an intrinsic atrix.

  Returns:
    Camera with updated intrinsics.
  """
  if intrinsic_matrix.shape[-2:] != (3, 3):
    raise ValueError(
        'Last two axes of intrinsic_matrix must have shape (3, 3).'
    )

  focal_length_x = intrinsic_matrix[..., 0, 0]
  focal_length_y = intrinsic_matrix[..., 1, 1]
  pixel_aspect_ratio = focal_length_y / focal_length_x
  principal_point_x = intrinsic_matrix[..., 0, 2]
  principal_point_y = intrinsic_matrix[..., 1, 2]
  return camera.replace(
      focal_length=focal_length_x,
      pixel_aspect_ratio=pixel_aspect_ratio,
      skew=intrinsic_matrix[..., 0, 1],
      principal_point=jnp.stack(
          [principal_point_x, principal_point_y], axis=-1
      ),
  )


def invert_radial_distortion_coefficients(
    radial_distortion: jnp.ndarray,
) -> jnp.ndarray:
  """Inverts radial distortion parameters.

  This uses the exact inverse formula from Drap and Lefèvre (Sensors, 2016).

  Reference:
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4934233/

  Args:
    radial_distortion: The input radial distortion parameters.

  Returns:
    The inverted radial distortion parameters.
  """
  a1, a2, a3, a4 = jnp.split(radial_distortion, 4, axis=-1)
  b1 = -a1
  b2 = 3 * a1**2 - a2
  b3 = 8 * a1 * a2 - 12 * a1**3 - a3
  b4 = 55 * a1**4 + 10 * a1 * a3 - 55 * a1**2 * a2 + 5 * a2**2 - a4
  return jnp.concatenate([b1, b2, b3, b4], axis=-1)


def _radial_and_tangential_distort(
    x: jnp.ndarray,
    y: jnp.ndarray,
    radial_distortion: jnp.ndarray | None = None,
    tangential_distortion: jnp.ndarray | None = None,
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Computes the distorted pixel positions."""
  dx_radial, dy_radial = 0.0, 0.0
  dx_tangential, dy_tangential = 0.0, 0.0
  r2 = x * x + y * y

  if radial_distortion is not None:
    k1, k2, k3, k4 = radial_distortion
    radial_distortion = r2 * (k1 + r2 * (k2 + r2 * (k3 + r2 * k4)))
    dx_radial = x * radial_distortion
    dy_radial = y * radial_distortion

  if tangential_distortion is not None:
    p1, p2 = tangential_distortion
    dx_tangential = 2 * p1 * x * y + p2 * (r2 + 2 * x * x)
    dy_tangential = 2 * p2 * x * y + p1 * (r2 + 2 * y * y)

  return x + dx_radial + dx_tangential, y + dy_radial + dy_tangential


def _radial_and_tangential_undistort(
    x_distorted: jnp.ndarray,
    y_distorted: jnp.ndarray,
    radial_distortion: jnp.ndarray,
    tangential_distortion: jnp.ndarray,
    max_iterations: int = 3,
) -> tuple[jnp.ndarray, jnp.ndarray]:
  """Undistorts a point in image-space using an iterative optimization.

  This uses Newton-Raphson (related to Gauss-Newton) to find the undistorted
  pixel positiong by finding the roots of the equation `distort(x) - x_d = 0`.

  References:
    https://en.wikipedia.org/wiki/Newton%27s_method#k_variables,_k_functions

  Args:
    x_distorted: The distorted x pixel position.
    y_distorted: The distorted y pixel position.
    radial_distortion: (4,) The radial distortion parameters.
    tangential_distortion: (2,) The tangential distortion parameters.
    max_iterations: The number of iterations to run Newton's method.

  Returns:
    A tuple containing the undistorted x and y positions.
  """

  def f(xy, xy_distorted):
    x_distorted_pred, y_distorted_pred = _radial_and_tangential_distort(
        xy[..., 0],
        xy[..., 1],
        radial_distortion=radial_distortion,
        tangential_distortion=tangential_distortion,
    )
    xy_distorted_pred = jnp.stack([x_distorted_pred, y_distorted_pred], axis=-1)
    return xy_distorted_pred - xy_distorted

  batch_shape = x_distorted.shape
  xy_distorted = jnp.stack([x_distorted, y_distorted], axis=-1)
  xy_distorted = xy_distorted.reshape(-1, 2)

  # Run Newton's method.
  jac_fn = jax.jacfwd(f)
  xy = xy_distorted
  for _ in range(max_iterations):
    jac = jax.vmap(jac_fn)(xy, xy_distorted)
    # Solve the linear system: J(x_n)(x{n+1}-x_n) = -f(x_n).
    # residual = x_{n+1} - x_n, so x_{n+1} = residual + x_n.
    residual = jax.vmap(jnp.linalg.solve)(jac, -f(xy, xy_distorted))
    xy = residual + xy

  # Reshape back to the original shape.
  xy = xy.reshape(*batch_shape, 2)
  return xy[..., 0], xy[..., 1]


@jax.jit
def project(camera: Camera, points: jnp.ndarray) -> jnp.ndarray:
  """Projects world-space 3D points (x,y,z) to pixel positions (x,y).

  Args:
    camera: The camera to project with.
    points: (..., 3) The 3D world-space points to project.

  Returns:
    A (..., 2) array containing the projected pixel locations.
  """
  eps = jnp.finfo(points.dtype).eps
  batch_shape = points.shape[:-1]
  local_points = world_points_to_local_points(camera, points)

  local_x = local_points[..., 0]
  local_y = local_points[..., 1]
  local_z = local_points[..., 2]

  # Get normalized local pixel positions.
  x = local_x / local_z
  y = local_y / local_z
  if camera.projection_type is ProjectionType.FISHEYE:
    r = jnp.sqrt(local_x ** 2 + local_y ** 2)
    theta = jnp.arctan2(r, local_z)
    fisheye_x = theta / r * local_x
    fisheye_y = theta / r * local_y
    # When theta is small it is approximately the same as perspective.
    x = jnp.where(theta < eps, x, fisheye_x)
    y = jnp.where(theta < eps, y, fisheye_y)

  x, y = _distort_local_pixels(camera, x, y)

  # Map the distorted ray to the image plane and return the depth.
  pixel_x = camera.focal_length * x + camera.skew * y + camera.principal_point_x
  pixel_y = (
      camera.focal_length * camera.pixel_aspect_ratio * y +
      camera.principal_point_y)

  pixels = jnp.stack([pixel_x, pixel_y], axis=-1)
  return pixels.reshape((*batch_shape, 2))


def look_at(
    camera: Camera,
    eye: jnp.ndarray,
    center: jnp.ndarray,
    world_up: jnp.ndarray,
    camera_convention: str = 'opencv',
) -> Camera:
  """Applies a look-at transform to the given camera.

  Functionality mimes gluLookAt (third_party/GL/glu/include/GLU/glu.h).

  Args:
    camera: The camera to move.
    eye: (3,) The position of the camera.
    center: (3,) The position the camera will "look at".
    world_up: (3,) The world's up direction.
    camera_convention: one of 'opengl' or 'opencv'. In 'opengl' convention, the
      camera aims down the -Z axis and the viewport origin is lower-left. In
      'opencv' convention, the camera aims down the +Z axis and the viewport
      origin is upper-left.

  Returns:
    A copy of the camera but with the look-at transform applied.
  """
  # NOTE: The computation before conversion happens in OpenGL coordinates.
  vector_degeneracy_cutoff = 1e-6
  forward = center - eye
  forward_norm = jnp.linalg.norm(forward)

  try:
    if forward_norm < vector_degeneracy_cutoff:
      raise ValueError(
          'Camera matrix is degenerate because eye and center are close.'
      )
  except jax.errors.ConcretizationTypeError:
    pass

  forward = forward / forward_norm

  to_side = jnp.cross(forward, world_up)
  to_side_norm = jnp.linalg.norm(to_side)
  try:
    if to_side_norm < vector_degeneracy_cutoff:
      raise ValueError(
          'Camera matrix is degenerate because up and gaze are close '
          'or up is degenerate.'
      )
  except jax.errors.ConcretizationTypeError:
    pass

  to_side = to_side / to_side_norm
  cam_up = jnp.cross(to_side, forward)

  # Make a 3x3 rotation matrix:
  view_rotation = jnp.vstack([to_side, cam_up, -forward])
  # Set the upper 3x3 of a 4x4 identity matrix:
  view_rotation = jnp.eye(4).at[:3, :3].set(view_rotation)

  # Make a 4x4 translation matrix:
  view_translation = jnp.eye(4).at[:3, 3].set(-eye)

  cam_from_world = math.matmul(view_rotation, view_translation)

  if camera_convention == 'opencv':
    # Equal to left multiplication by np.diag([1.0, -1.0, -1.0, 1.0])
    cam_from_world = cam_from_world.at[1:3, :].multiply(-1.0)
  elif camera_convention != 'opengl':
    raise ValueError(f'Unknown camera convention {camera_convention}')

  return update_world_to_camera_matrix(camera, cam_from_world)


def get_pixel_centers(image_width: int, image_height: int) -> jnp.ndarray:
  """Returns the pixel centers."""
  xx, yy = jnp.meshgrid(
      jnp.arange(image_width, dtype=jnp.float32),
      jnp.arange(image_height, dtype=jnp.float32))
  return jnp.stack([xx, yy], axis=-1) + 0.5


def scale(camera: Camera, amount: float | jnp.ndarray) -> Camera:
  """Scales the camera by the given amount."""
  # First round the image size to a round number.
  new_image_size = jnp.round(camera.image_size * amount)
  return scale_to_image_size(camera, new_image_size)


def scale_to_image_size(camera: Camera, image_size: jnp.ndarray) -> Camera:
  """Scales the camera to the given image size."""
  scale_xy = image_size / camera.image_size
  scale_x = scale_xy[..., 0]
  scale_y = scale_xy[..., 1]
  return camera.replace(
      focal_length=camera.focal_length * scale_x,
      pixel_aspect_ratio=scale_y / scale_x * camera.pixel_aspect_ratio,
      principal_point=camera.principal_point * scale_xy,
      skew=camera.skew * scale_x,
      image_size=image_size,
  )


def crop(
    camera: Camera,
    left: int | jnp.ndarray = 0,
    right: int | jnp.ndarray = 0,
    top: int | jnp.ndarray = 0,
    bottom: int | jnp.ndarray = 0,
) -> Camera:
  """Returns a copy of the camera with adjusted image bounds.

  NOTE: This function does not perform bounds checking.

  Args:
    camera: Input camera.
    left: Number of pixels by which to reduce (or augment, if negative) the left
      boundary of the image domain extent.
    right: Number of cropped pixels for the right image boundary.
    top: Number of cropped pixels for the top image boundary.
    bottom: Number of cropped pixels for the bottom image boundary.

  Returns:
    A camera with adjusted image dimensions. The focal length is unchanged, and
    the principal point is updated to preserve the original principal axis.
  """
  crop_left_top = jnp.array([left, top])
  crop_right_bottom = jnp.array([right, bottom])
  new_image_size = camera.image_size - crop_left_top - crop_right_bottom
  new_principal_point = camera.principal_point - crop_left_top

  return camera.replace(
      principal_point=new_principal_point,
      image_size=new_image_size,
  )


def concatenate(cameras: Sequence[Camera], axis: int = 0) -> Camera:
  """Concatenates two batched cameras.

  This function operates primarily on _batched_ cameras. Either all cameras
  must be batched, or all cameras must be unbatched. In the latter case a single
  batch dimension will be created, similar to when concatenating scalars in
  numpy.

  If some attributes are unbatched, they will be expanded and broadcasted to the
  batch shape.

  Args:
    cameras: A list of cameras to concatenate.
    axis: The batch axis to concatenate across.

  Raises:
    ValueError: If no cameras are given.
    TypeError: If the batch axes do not match.

  Returns:
    The concatenated camera.
  """

  if not cameras:
    raise ValueError('Need at least one camera to concatenate.')

  if not all(c.ndim == cameras[0].ndim for c in cameras):
    dimensions = ', '.join([str(c.shape) for c in cameras])
    raise TypeError(
        f'Cannot concatenate cameras with different numbers of dimensions'
        f': got {dimensions}')

  # Expand unbatched cameras.
  cameras = [(c if c.shape else jax.tree.map(
      functools.partial(jnp.expand_dims, axis=0), c)) for c in cameras]

  cls = cameras[0].__class__
  kwargs = {}
  for field in dataclasses.fields(cls):
    # For non-field values pass the value of the first camera assuming they are
    # all the same.
    first_value = getattr(cameras[0], field.name)
    if not field.metadata['pytree_node'] or first_value is None:
      kwargs[field.name] = first_value
      continue

    unbatched_shape = field.metadata['unbatched_shape']

    def _expand_and_broadcast(c):
      # pylint: disable=cell-var-from-loop
      v = jnp.asarray(getattr(c, field.name))
      ndim = v.ndim - len(unbatched_shape)
      v = jnp.expand_dims(v, axis=tuple(range(len(c.shape) - ndim)))
      v = jnp.broadcast_to(v, shape=(*c.shape, *unbatched_shape))
      return v

    values = [_expand_and_broadcast(c) for c in cameras]
    kwargs[field.name] = jnp.concatenate(values, axis=axis)

  return cls(**kwargs)


def transform(
    camera: Camera,
    scale: jnp.ndarray,  # pylint: disable=redefined-outer-name
    rotation: jnp.ndarray,
    translation: jnp.ndarray,
) -> Camera:
  """Applies a similarity transform to the camera.

  The scale and rotation are applied first followed by the translation. I.e.,
  if the translation is in the input coordinates, it should be rotated and
  scaled prior to being passed to this function.

  The new position becomes `scale * rotation @ position + translation`.
  The new orientation becomes `orientation @ rotation.T`.

  Args:
    camera: The camera to transform.
    scale: (1,) The scale of the transform.
    rotation: (3, 3) The rotation of the transform.
    translation: (3,) The translation of the transform.

  Returns:
    The transformed camera.
  """
  new_orientation = math.matmul(camera.orientation, rotation.T)
  new_position = math.transform_point(
      camera.position, scale=scale, rotation=rotation, translation=translation
  )
  camera = camera.replace(orientation=new_orientation, position=new_position)
  return camera


def relative_transform(reference: Camera, target: Camera) -> jnp.ndarray:
  """Computes the transform from reference camera to the target camera.

  The transform relativizes the camera such that the target cameras are in the
  reference camera's reference frame. I.e., the reference camera's camera
  coordinate frame becomes the world coordinate frame.

  Args:
    reference: The reference camera.
    target: The target camera.

  Returns:
    The relative transform between the target and reference camera. This matrix
    takes points in the reference camera's coordinates and transforms them to
    be in the target camera's coordinates.
  """
  return math.matmul(
      target.world_to_camera_matrix, reference.camera_to_world_matrix
  )


def essential_matrix(reference: Camera, target: Camera) -> jnp.ndarray:
  """Computes the essential matrix between two cameras.

  The essential matrix will be computed based on the transformation from the
  reference to the target camera.

  References:
    https://en.wikipedia.org/wiki/Essential_matrix

  Args:
    reference: The reference camera.
    target: The target camera.

  Returns:
    A (3, 3) matrix containing the essential matrix.
  """
  rel_transform = relative_transform(reference, target)
  rel_target = update_world_to_camera_matrix(target, rel_transform)
  return math.matmul(math.skew(rel_target.translation), rel_target.orientation)


def fundamental_matrix(reference: Camera, target: Camera) -> jnp.ndarray:
  """Computes the fundamental matrix between two cameras.

  References:
    https://en.wikipedia.org/wiki/Fundamental_matrix_(computer_vision)

  Args:
    reference: The reference camera.
    target: The target camera.

  Returns:
    A (3, 3) matrix containing the fundamental matrix. When this matrix is left
    multiplied to a homogeneous pixel coordinate in the reference camera, it
    results in a homogeneous line equation containing the corresponding epipolar
    line in the target camera.
  """
  essential_mat = essential_matrix(reference, target)
  left = jnp.linalg.inv(target.intrinsic_matrix).T
  right = jnp.linalg.inv(reference.intrinsic_matrix)
  return math.matmul(left, math.matmul(essential_mat, right))



# Copyright 2024 The jaxcam Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The public interface camera I/O."""

# pylint: disable=g-multiple-import,g-importing-member,useless-import-alias,unused-import
# Note: import <name> as <name> is required for names to be exported.
# See PEP 484 & https://github.com/google/jax/issues/7570
from jaxcam._src.io import (
    from_nerfies_json_file as from_nerfies_json_file,
    to_nerfies_json_file as to_nerfies_json_file,
)
